# -*- coding: utf-8 -*-
"""
authentication handlers base module.
"""

from abc import abstractmethod

import pyrin.security.token.services as token_services
import pyrin.security.session.services as session_services
import pyrin.utils.misc as misc_utils

from pyrin.core.globals import _
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.enumerations import TokenTypeEnum
from pyrin.security.authentication.interface import AbstractAuthenticatorBase
from pyrin.security.authentication.handlers.exceptions import RefreshTokenRequiredError, \
    AuthenticatorNameIsRequiredError, AccessTokenRequiredError, InvalidUserError, \
    AccessAndRefreshTokensDoesNotBelongToSameUserError, InvalidAccessTokenError, \
    InvalidTokenAuthenticatorError, InvalidRefreshTokenError, UserCredentialsRevokedError, \
    InvalidUserIdentityError, ProvidedUsernameOrPasswordAreIncorrect


class AuthenticatorBase(AbstractAuthenticatorBase):
    """
    authenticator base class.

    all application authenticators must be subclassed from this.
    """

    # each subclass must set an authenticator name in this attribute.
    _name = None

    def __init__(self, *args, **options):
        """
        initializes and instance of AuthenticatorBase.

        :raises AuthenticatorNameIsRequiredError: authenticator name is required error.
        """

        super().__init__()

        if not self._name or self._name.isspace():
            raise AuthenticatorNameIsRequiredError('Authenticator [{instance}] does not '
                                                   'have a name.'.format(instance=self))

    def _get_info(self, payload, **options):
        """
        gets the info of given payload to be set in current request.

        :param dict | str payload: credential payload.

        :rtype: dict
        """

        result = dict()
        extra_info = self._get_extra_info(payload, **options)
        if extra_info:
            result.update(extra_info)

        return result

    def _get_extra_info(self, payload, **options):
        """
        gets the info of given payload to be set in current request.

        it could be None if no extra info must be set in current request.

        :param dict | str payload: credential payload.

        :rtype: dict
        """

        return None

    def _get_user_info(self, user, **options):
        """
        gets the info of given user to be set in current request.

        it could be None if no extra info must be set in current request.

        :param BaseEntity | ROW_RESULT user: user entity to get its info.

        :rtype: object | dict
        """

        return None

    def _get_custom_component_key(self, user, *args, **options):
        """
        gets custom component key for given user if required to be set in current request.

        this method could be overridden in subclasses if required.

        :param BaseEntity | ROW_RESULT user: authenticated user entity.

        :rtype: object
        """

        return None

    def _set_custom_component_key(self, value):
        """
        sets the provided value as custom component key into current request.

        :param object value: value to be pushed as component custom key.

        :raises InvalidComponentCustomKeyError: invalid component custom key error.
        """

        session_services.set_component_custom_key(value)

    def _pre_authenticate(self, *payloads, **options):
        """
        pre-authenticates the user with given credentials.

        this method could be overridden in subclasses if pre-authentication is required.
        it must raise an `AuthenticationFailedError` if pre-authentication failed.

        :param dict | str payloads: credential payloads.
                                    it is usually an access token, refresh
                                    token or a session identifier.
                                    it can be multiple items if required.
        """
        pass

    def _authenticate(self, *payloads, **options):
        """
        authenticates the user with given credentials.

        :param dict | str payloads: credential payloads.
                                    it is usually an access token, refresh
                                    token or a session identifier.
                                    it can be multiple items if required.

        :raises UserCredentialsRevokedError: user credentials revoked error.
        :raises AuthenticationFailedError: authentication failed error.
        :raises InvalidUserError: invalid user error.
        :raises InvalidUserIdentityError: invalid user identity error.
        """

        if self._is_revoked(*payloads, **options):
            raise UserCredentialsRevokedError(_('User credentials are revoked.'))

        self._pre_authenticate(*payloads, **options)
        user_payload = self._get_user_related_payload(*payloads, **options)
        user = self._get_user(user_payload, **options)
        if not user:
            raise InvalidUserError('User could not be None.')

        identity = self._get_user_identity(user, **options)
        if not identity:
            raise InvalidUserIdentityError('User identity could not be None.')

        info = self._get_info(user_payload, **options)
        user_info = self._get_user_info(user, **options)
        if user_info:
            info.update(user_info)

        session_services.set_current_user(identity, info)
        custom_component_key = self._get_custom_component_key(user, user_payload, **options)
        if custom_component_key is not None:
            self._set_custom_component_key(custom_component_key)

    def _is_revoked(self, *payloads, **options):
        """
        gets a value indicating that given payloads are revoked.

        if you want to implement credential revocation, you must implement
        this method. otherwise it will always return False.

        :param dict | str payloads: credential payloads.
                                    it is usually an access token, refresh
                                    token or a session identifier.
                                    it can be multiple items if required.

        :rtype: bool
        """

        return False

    def _persist_payloads(self, user, *payloads, **options):
        """
        persists given payloads for user login.

        this is needed if you want to use credentials revocation.
        if you do not need this feature, you can leave this method unimplemented.

        :param BaseEntity | ROW_RESULT user: user to persist its generated credentials.

        :param dict | str payloads: credential payloads.
                                    it is usually an access token, refresh
                                    token or a session identifier.
                                    it can be multiple items if required.
        """
        pass

    def _persist_login(self, user, **options):
        """
        persists required login data for given user.

        this method is intended to be overridden in subclasses to perform
        custom actions. for example updating the last login datetime of user.

        :param BaseEntity | ROW_RESULT user: user to persist its login data.
        """
        pass

    def _generate_credentials(self, user, **options):
        """
        generates the required credentials for given user for a successful login.

        :param BaseEntity | ROW_RESULT user: user to be logged in.

        :returns: required credentials to be returned to client.
        """

        credentials = self._create_credentials(user, **options)
        credentials = misc_utils.make_iterable(credentials, tuple)
        payloads = self._get_payloads(*credentials, **options)
        payloads = misc_utils.make_iterable(payloads, tuple)
        self._persist_payloads(user, *payloads, **options)
        return self._get_client_credentials(*credentials, **options)

    def authenticate(self, request, **options):
        """
        authenticates the user for given request.

        :param CoreRequest request: current request object.

        :raises AuthenticationFailedError: authentication failed error.
        """

        credentials = self._get_credentials(request, **options)
        credentials = misc_utils.make_iterable(credentials, tuple)
        payloads = self._get_payloads(*credentials, **options)
        payloads = misc_utils.make_iterable(payloads, tuple)
        self._authenticate(*payloads, **options)

    def is_fresh(self):
        """
        gets a value indicating that current request has a fresh authentication.

        fresh authentication means an authentication which is done by providing
        user credentials to server.
        if you do not want to implement such a concept in your authenticator,
        you can leave this method unimplemented and it will always return False.

        :rtype: bool
        """

        return False

    def login(self, username, password, **options):
        """
        logs in a user with given info and stores/generates the relevant credentials.

        it may return the required credentials if they must be sent to client.

        :param str username: username.
        :param str password: password.

        :raises ProvidedUsernameOrPasswordAreIncorrect: provided username or
                                                        password are incorrect.

        :returns: required credentials to be returned to client.
        """

        user = self._get_login_user(username, password, **options)
        if user is None:
            raise ProvidedUsernameOrPasswordAreIncorrect(_('The provided username or '
                                                           'password are incorrect.'))

        self._validate_login(user, **options)
        credentials = self._generate_credentials(user, **options)
        self._persist_login(user, **options)
        return credentials

    def logout(self, **options):
        """
        logouts the current user and clears its relevant credentials.
        """

        user = session_services.get_current_user()
        credentials = self._get_credentials(session_services.get_current_request(), **options)
        credentials = misc_utils.make_iterable(credentials, tuple)
        payloads = self._get_payloads(*credentials, **options)
        payloads = misc_utils.make_iterable(payloads, tuple)
        self._revoke(user, *payloads, **options)

    @abstractmethod
    def _create_credentials(self, user, **options):
        """
        creates the required credentials for given user for a successful login.

        :param BaseEntity | ROW_RESULT user: user to be logged in.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: created credentials.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_client_credentials(self, *credentials, **options):
        """
        gets the credentials that must be returned to client.

        :param str credentials: user credentials.
                                it is usually the contents of
                                authorization or cookie headers.
                                it can be multiple items if required.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: required credentials to be returned to client.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _revoke(self, *payloads, **options):
        """
        revokes the given payloads for user logout.

        :param dict | str payloads: credential payloads.
                                    it is usually an access token, refresh
                                    token or a session identifier.
                                    it can be multiple items if required.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_login_user(self, username, password, **options):
        """
        gets the related user entity to given inputs for logging in.

        it must return None if no user found.

        :param str username: username.
        :param str password: password.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: BaseEntity | ROW_RESULT
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _validate_login(self, user, **options):
        """
        validates that given user can actually logged in.

        if user should not be logged in it must raise a relevant authorization failed error.

        :param BaseEntity | ROW_RESULT user: user to be logged in.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_payloads(self, *credentials, **options):
        """
        gets the required payloads from given credentials.

        it can return multiple items if required.

        :param str credentials: user credentials.
                                it is usually the contents of
                                authorization or cookie headers.
                                it can be multiple items if required.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: tuple[dict | str] | dict | str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_credentials(self, request, **options):
        """
        gets the required credentials from given request.

        credentials are usually the contents of authorization or cookie headers.
        it can return multiple items if required.

        :param CoreRequest request: current request object.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: tuple[str] | str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_user_related_payload(self, *payloads, **options):
        """
        gets a single value as user related payload to be used to fetch related user.

        :param dict | str payloads: credential payloads.
                                    it is usually an access token, refresh
                                    token or a session identifier.
                                    it can be multiple items if required.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict | str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_user_identity(self, user, **options):
        """
        gets the identity of given user to be set in current request.

        the identity is normally the primary key of user entity.
        but it could be a dict of multiple values if required.

        :param BaseEntity | ROW_RESULT user: user entity to get its identity.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object | dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_user(self, payload, **options):
        """
        gets the user entity from given inputs.

        this method must return a user on success or raise an error
        if it can not fetch related user to given payload.

        :param dict | str payload: credential payload.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: BaseEntity | ROW_RESULT
        """

        raise CoreNotImplementedError()

    @property
    def name(self):
        """
        gets the name of this authenticator.

        :rtype: str
        """

        return self._name


class TokenAuthenticatorBase(AuthenticatorBase):
    """
    token authenticator base class.

    all application token authenticators must be subclassed from this.
    """

    # header name to read access token from it.
    ACCESS_TOKEN_HOLDER = 'Authorization'

    # cookie or header name to read refresh token from it.
    REFRESH_TOKEN_HOLDER = 'Refresh-Auth'

    # a key name to hold user identity in token payloads.
    USER_IDENTITY_HOLDER = 'sub'

    # a key name to hold authenticator name in token payloads.
    AUTHENTICATOR_HOLDER = 'auth'

    # specifies that this authenticator requires refresh token.
    _refresh_token = True

    # specifies that refresh token must be read from a cookie.
    # if set to False, it will be read from a normal header.
    _refresh_token_in_cookie = True

    def _get_token_payload(self, token, **options):
        """
        gets the payload of given token.

        :param str token: token.

        :returns: tuple[dict header, dict payload]
        :rtype: tuple[dict, dict]
        """

        header = token_services.get_unverified_header(token, **options)
        payload = token_services.get_payload(token, **options)
        return header, payload

    def _get_access_token_payload(self, access_token, **options):
        """
        gets the given access token payload.

        :param str access_token: access token.

        :raises AccessTokenRequiredError: access token required error.

        :rtype: dict
        """

        if not access_token:
            raise AccessTokenRequiredError(_('Access token is required for authentication.'))

        header, payload = self._get_token_payload(access_token, **options)
        self._validate_access_token(header, payload, **options)
        return payload

    def _get_refresh_token_payload(self, refresh_token, **options):
        """
        gets the given refresh token payload.

        :param str refresh_token: refresh token.

        :raises RefreshTokenRequiredError: refresh token required error.

        :rtype: dict
        """

        if not refresh_token:
            raise RefreshTokenRequiredError(_('Refresh token is required for authentication.'))

        header, payload = self._get_token_payload(refresh_token, **options)
        self._validate_refresh_token(header, payload, **options)
        return payload

    def _get_payloads(self, access_token, refresh_token, **options):
        """
        gets the required payloads from given credentials.

        :param str access_token: access token.
        :param str refresh_token: refresh token.

        :returns: tuple[dict access_token_payload, dict refresh_token_payload]
        :rtype: tuple[dict]
        """

        access_token_payload = self._get_access_token_payload(access_token, **options)
        refresh_token_payload = None
        if self._refresh_token:
            refresh_token_payload = self._get_refresh_token_payload(refresh_token, **options)

        return access_token_payload, refresh_token_payload

    def _get_access_token_credential(self, request):
        """
        gets access token from given request.

        :param CoreRequest request: current request object.

        :rtype: str
        """

        return request.headers.get(self.ACCESS_TOKEN_HOLDER)

    def _get_refresh_token_credential(self, request):
        """
        gets refresh token from given request.

        :param CoreRequest request: current request object.

        :rtype: str
        """

        if self._refresh_token_in_cookie:
            return request.cookies.get(self.REFRESH_TOKEN_HOLDER)

        return request.headers.get(self.REFRESH_TOKEN_HOLDER)

    def _get_credentials(self, request, **options):
        """
        gets the required credentials from given request.

        credentials are usually the contents of authorization or cookie headers.
        it can return multiple items if required.

        :param CoreRequest request: current request object.

        :returns: tuple[str access_token, str refresh_token]
        :rtype: tuple[str, str]
        """

        access_token = self._get_access_token_credential(request)
        refresh_token = None
        if self._refresh_token:
            refresh_token = self._get_refresh_token_credential(request)

        return access_token, refresh_token

    def _validate_same_user(self, access_token_payload, refresh_token_payload, **options):
        """
        validates that both tokens are related to the same user.

        :param dict access_token_payload: access token payload.
        :param dict refresh_token_payload: refresh token payload.

        :raises AccessAndRefreshTokensDoesNotBelongToSameUserError: access and refresh tokens
                                                                    does not belong to same
                                                                    user error.
        """

        access_user = access_token_payload.get(self.USER_IDENTITY_HOLDER)
        refresh_user = refresh_token_payload.get(self.USER_IDENTITY_HOLDER)
        if access_user != refresh_user:
            raise AccessAndRefreshTokensDoesNotBelongToSameUserError(_('Provided access and '
                                                                       'refresh tokens does not '
                                                                       'belong to the same user.'))

    def _validate_access_token(self, header, payload, **options):
        """
        validates given header and payload of an access token.

        :param dict header: token header.
        :param dict payload: token payload.

        :raises InvalidAccessTokenError: invalid access token error.
        :raises InvalidTokenAuthenticatorError: invalid token authenticator error.
        """

        if not header or not payload or not payload.get(self.USER_IDENTITY_HOLDER) or \
                payload.get('type') != TokenTypeEnum.ACCESS:
            raise InvalidAccessTokenError(_('Provided access token is invalid.'))

        generator = payload.get(self.AUTHENTICATOR_HOLDER)
        if generator != self.name:
            raise InvalidTokenAuthenticatorError(_('This access token is generated using '
                                                   'another authenticator with name [{name}].')
                                                 .format(name=generator))

    def _validate_refresh_token(self, header, payload, **options):
        """
        validates given header and payload of a refresh token.

        :param dict header: token header.
        :param dict payload: token payload.

        :raises InvalidRefreshTokenError: invalid refresh token error.
        :raises InvalidTokenAuthenticatorError: invalid token authenticator error.
        """

        if not header or not payload or not payload.get(self.USER_IDENTITY_HOLDER) or \
                payload.get('type') != TokenTypeEnum.REFRESH:
            raise InvalidRefreshTokenError(_('Provided refresh token is invalid.'))

        generator = payload.get(self.AUTHENTICATOR_HOLDER)
        if generator != self.name:
            raise InvalidTokenAuthenticatorError(_('This refresh token is generated using '
                                                   'another authenticator with name [{name}].')
                                                 .format(name=generator))

    def _get_extra_info(self, payload, **options):
        """
        gets the info of given payload to be set in current request.

        :param dict payload: access token payload.

        :rtype: dict
        """

        return dict(is_fresh=payload.get('is_fresh') or False)

    def _pre_authenticate(self,  access_token_payload, refresh_token_payload, **options):
        """
        pre-authenticates the user with given credentials.

        :param dict access_token_payload: access token payload.
        :param dict refresh_token_payload: refresh token payload.

        :raises AccessAndRefreshTokensDoesNotBelongToSameUserError: access and refresh tokens
                                                                    does not belong to same
                                                                    user error.

        :raises UserCredentialsRevokedError: user credentials revoked error.
        """

        if self._refresh_token:
            self._validate_same_user(access_token_payload,
                                     refresh_token_payload, **options)

    def _get_user_related_payload(self, access_token_payload,
                                  refresh_token_payload, **options):
        """
        gets a single value as user related payload to be used to fetch related user.

        it simply returns the access token payload.

        :param dict access_token_payload: access token payload.
        :param dict refresh_token_payload: refresh token payload.

        :rtype: dict
        """

        return access_token_payload

    def _create_credentials(self, user, **options):
        """
        creates the required access and refresh tokens for given user for a successful login.

        :param BaseEntity | ROW_RESULT user: user to be logged in.

        :returns: tuple[str access_token, str refresh_token]
        :rtype: tuple[str, str]
        """

        access_claims = self._get_access_token_claims(user, **options) or {}
        refresh_claims = {}
        if self._refresh_token:
            refresh_claims = self._get_refresh_token_claims(user, **options) or {}

        common_claims = dict()
        common_claims[self.AUTHENTICATOR_HOLDER] = self.name
        common_claims[self.USER_IDENTITY_HOLDER] = self._get_user_identity(user, **options)
        access_claims.update(common_claims)
        if self._refresh_token:
            refresh_claims.update(common_claims)

        access_token = token_services.generate_access_token(access_claims, is_fresh=True)
        refresh_token = None
        if self._refresh_token:
            refresh_token = token_services.generate_refresh_token(refresh_claims)

        return access_token, refresh_token

    def _get_access_token_claims(self, user, **options):
        """
        gets a dict of all claims that must be added to access token payload.

        this method could be overridden in subclasses.

        :param BaseEntity | ROW_RESULT user: user to be logged in.

        :rtype: dict
        """

        return {}

    def _get_refresh_token_claims(self, user, **options):
        """
        gets a dict of all claims that must be added to refresh token payload.

        this method could be overridden in subclasses.

        :param BaseEntity | ROW_RESULT user: user to be logged in.

        :rtype: dict
        """

        return {}

    def _get_client_credentials(self, access_token, refresh_token, **options):
        """
        gets the credentials that must be returned to client.

        :param str access_token: access token.
        :param str refresh_token: refresh token.

        :returns: dict(str access_token) | dict(str access_token, str refresh_token)
        :rtype: dict
        """

        result = dict(access_token=access_token)
        if self._refresh_token:
            if self._refresh_token_in_cookie:
                session_services.set_response_cookie(self.REFRESH_TOKEN_HOLDER, refresh_token)
            else:
                result.update(refresh_token=refresh_token)

        return result

    def is_fresh(self):
        """
        gets a value indicating that current request has a fresh authentication.

        fresh authentication means an authentication which is done by providing
        user credentials to server.

        :rtype: bool
        """

        user_info = session_services.get_current_user_info()
        return user_info and user_info.get('is_fresh') is True



# -*- coding: utf-8 -*-
"""
security enumerations module.
"""

from pyrin.core.enumerations import CoreEnum, EnumMember


class TokenTypeEnum(CoreEnum):
    """
    token type enum.
    """

    ACCESS = EnumMember('access', 'Access')
    REFRESH = EnumMember('refresh', 'Refresh')


class InternalAuthenticatorEnum(CoreEnum):
    """
    internal authenticator enum.
    """

    ADMIN = EnumMember('admin', 'Admin')
    AUDIT = EnumMember('audit', 'Audit')
    SWAGGER = EnumMember('swagger', 'Swagger')



# -*- coding: utf-8 -*-
"""
admin decorators module.
"""

import pyrin.admin.services as admin_services


def admin(*args, **kwargs):
    """
    decorator to register an admin page.

    :param object args: admin class constructor arguments.
    :param object kwargs: admin class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           admin page with the same register name or the same entity,
                           replace it with the new one, otherwise raise
                           an error. defaults to False.

    :raises InvalidAdminPageTypeError: invalid admin page type error.
    :raises DuplicatedAdminPageError: duplicated admin page error.

    :returns: admin class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available admin pages.

        :param type cls: admin class.

        :returns: admin class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        admin_services.register(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
authentication handlers internal module.
"""

import pyrin.users.internal.services as internal_user_services

from pyrin.core.globals import _
from pyrin.users.internal.models import InternalUserEntity
from pyrin.security.exceptions import UserIsNotActiveError
from pyrin.security.authentication.handlers.base import TokenAuthenticatorBase


class InternalTokenAuthenticator(TokenAuthenticatorBase):
    """
    internal token authenticator class.

    all application internal token authenticators must be subclassed from this.
    for example admin, swagger or audit authenticators.
    """

    def _get_user_info(self, user, **options):
        """
        gets the info of given user to be set in current request.

        it could be None if no extra info must be set in current request.

        :param ROW_RESULT user: internal user entity to get its info.

        :rtype: dict
        """

        result = dict(fullname=user.fullname,
                      is_active=user.is_active,
                      is_superuser=user.is_superuser)

        extra_info = self._get_extra_user_info(user, **options)
        if extra_info:
            result.update(extra_info)

        return result

    def _get_user_identity(self, user, **options):
        """
        gets the identity of given user to be set in current request.

        the identity is normally the primary key of user entity.
        but it could be a dict of multiple values if required.

        :param ROW_RESULT user: internal user entity to get its identity.

        :rtype: int
        """

        return user.id

    def _get_user(self, payload, **options):
        """
        gets the internal user from given access token payload.

        this method must return an internal user on success or raise an
        exception if it can not fetch related user to given access token payload.

        :param dict payload: access token payload.

        :rtype: ROW_RESULT
        """

        user_id = payload.get(self.USER_IDENTITY_HOLDER)
        return internal_user_services.get(user_id,
                                          InternalUserEntity.id,
                                          InternalUserEntity.is_active,
                                          InternalUserEntity.is_superuser,
                                          InternalUserEntity.swagger_access,
                                          InternalUserEntity.audit_access,
                                          InternalUserEntity.admin_access,
                                          InternalUserEntity.fullname)

    def _get_extra_user_info(self, user, **options):
        """
        gets extra user info for given internal user if required.

        this method is intended to be overridden in subclasses.

        :param ROW_RESULT user: internal user to get its extra info.

        :rtype: dict
        """

        return None

    def _is_revoked(self, access_token_payload, refresh_token_payload, **options):
        """
        gets a value indicating that given payloads are revoked.

        :param dict access_token_payload: access token payload.
        :param dict refresh_token_payload: refresh token payload.

        :rtype: bool
        """

        # query on tokens

    def _persist_payloads(self, user, access_token_payload,
                          refresh_token_payload, **options):
        """
        persists given payloads for internal user login.

        :param ROW_RESULT user: internal user to persist its generated credentials.
        :param dict access_token_payload: access token payload.
        :param dict refresh_token_payload: refresh token payload.
        """
        pass

    def _revoke(self, access_token_payload, refresh_token_payload, **options):
        """
        revokes the given payloads for internal user logout.

        :param dict access_token_payload: access token payload.
        :param dict refresh_token_payload: refresh token payload.
        """

    def _get_login_user(self, username, password, **options):
        """
        gets the related internal user entity to given inputs for logging in.

        it must return None if no internal user found.

        :param str username: username.
        :param str password: password.

        :rtype: ROW_RESULT
        """

        return internal_user_services.get_login_user(username, password,
                                                     InternalUserEntity.id,
                                                     InternalUserEntity.is_active,
                                                     InternalUserEntity.is_superuser,
                                                     InternalUserEntity.swagger_access,
                                                     InternalUserEntity.audit_access,
                                                     InternalUserEntity.admin_access,
                                                     **options)

    def _validate_login(self, user, **options):
        """
        validates that given internal user can actually logged in.

        :param ROW_RESULT user: internal user to be logged in.

        :raises UserIsNotActiveError: user is not active error.
        """

        if user.is_active is not True:
            raise UserIsNotActiveError(_('Your user is not active. If you think that this '
                                         'is a mistake, please contact the support team.'))

        if user.is_superuser is not True:
            self._validate_access(user, **options)

    def _validate_access(self, user, **options):
        """
        validates that given user has access for login.

        it could be implemented in subclasses to check for access.

        :param ROW_RESULT user: internal user to be logged in.
        """
        pass

    def _persist_login(self, user, **options):
        """
        persists required login data for given user.

        :param BaseEntity | ROW_RESULT user: user to persist its login data.
        """

        internal_user_services.update_last_login_at(user.id, **options)



# -*- coding: utf-8 -*-
"""
authorization interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class AuthorizerSingletonMeta(MultiSingletonMeta):
    """
    authorizer singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractAuthorizerBase(CoreObject, metaclass=AuthorizerSingletonMeta):
    """
    abstract authorizer base class.
    """

    @abstractmethod
    def authorize(self, user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of the specified
        permissions, an error will be raised.

        :param user: user identity to authorize permissions for.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :keyword dict user_info: user info to be used for authorization.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def is_superuser(self):
        """
        gets a value indicating that the current user is superuser.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this authorizer.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
users internal admin module.
"""

import pyrin.users.internal.services as internal_user_services

from pyrin.admin.decorators import admin
from pyrin.admin.page.base import AdminPage
from pyrin.admin.page.decorators import link
from pyrin.users.internal.models import InternalUserEntity
from pyrin.security.authentication.models import InternalSessionEntity


@admin()
class InternalUserAdmin(AdminPage):
    entity = InternalUserEntity
    register_name = 'internal-users'
    name = 'Internal User'
    category = 'USERS'
    create_service = internal_user_services.create
    update_service = internal_user_services.update
    extra_data_fields = dict(password=str, confirm_password=str)
    list_search_fields = (InternalUserEntity.email,)
    list_fields = (InternalUserEntity.id, InternalUserEntity.username,
                   InternalUserEntity.fullname, InternalUserEntity.is_active,
                   InternalUserEntity.is_superuser, InternalUserEntity.admin_access,
                   InternalUserEntity.audit_access, InternalUserEntity.swagger_access,
                   InternalUserEntity.mobile, InternalUserEntity.gender,
                   InternalUserEntity.last_login_at, InternalUserEntity.created_at,
                   InternalUserEntity.modified_at, 'sessions')

    @link
    def sessions(self, row):
        """
        gets all related sessions to this user.

        :param ROW_RESULT row: user row.

        :rtype: dict
        """

        return self._get_link_info('Sessions', InternalSessionEntity, user_id=row.id)


@admin()
class InternalSessionAdmin(AdminPage):
    entity = InternalSessionEntity
    register_name = 'internal-sessions'
    name = 'Internal Session'
    category = 'AUTHENTICATION'
    list_fields = (InternalSessionEntity.id, InternalSessionEntity.user_id,
                   InternalUserEntity.fullname, InternalSessionEntity.type,
                   InternalSessionEntity.authenticator, InternalSessionEntity.is_revoked,
                   InternalSessionEntity.created_at, InternalSessionEntity.modified_at)

    def _perform_joins(self, query, **options):
        query = query.join(InternalUserEntity,
                           InternalSessionEntity.user_id == InternalUserEntity.id)
        return query



# -*- coding: utf-8 -*-
"""
authentication manager module.
"""

import pyrin.validator.services as validator_services
import pyrin.application.services as application_services
import pyrin.security.session.services as session_services
import pyrin.configuration.services as config_services
import pyrin.utils.dictionary as dict_utils

from pyrin.core.structs import Manager, Context
from pyrin.utils.custom_print import print_warning
from pyrin.api.router.handlers.protected import ProtectedRoute
from pyrin.security.authentication import AuthenticationPackage
from pyrin.security.authentication.interface import AbstractAuthenticatorBase
from pyrin.security.exceptions import AuthenticationFailedError
from pyrin.security.authentication.exceptions import InvalidAuthenticatorTypeError, \
    DuplicatedAuthenticatorError, AuthenticatorNotFoundError


class AuthenticationManager(Manager):
    """
    authentication manager class.
    """

    package_class = AuthenticationPackage

    def __init__(self):
        """
        initializes an instance of AuthenticationManager.
        """

        super().__init__()

        # a dictionary containing information of registered authenticators.
        # example: dict(str name: AbstractAuthenticatorBase instance)
        self._authenticators = Context()
        self._default_authenticator = config_services.get_active('authentication',
                                                                 'default_authenticator')

        # a dictionary containing all rule based authenticators in the form of:
        # dict(str url_rule: str name)
        # for example:
        # {'/api/customers': 'customer',
        #  '/api/sellers': 'seller',
        #  '/api': 'api'}
        self._rule_based_authenticators = self._load_rule_based_authenticators()

    def _load_rule_based_authenticators(self):
        """
        loads rule based authenticators from `authentication` config store.

        we sort rule based authenticators by length of their name in descending order.
        this is required to be able to detect the closest match to each url rule.

        :rtype: dict
        """

        authenticators = config_services.get_active('authentication',
                                                    'rule_based_authenticators')
        return dict_utils.sort_by_key_length(authenticators, reverse=True)

    def get_relevant_authenticator_name(self, url):
        """
        gets the relevant authenticator name for given url rule.

        it looks for relevant authenticator in the following order:

        1. looking for it in rule based authenticators.
        2. getting default authenticator for it.

        it may return None if no relevant authenticator could be found.

        :param str url: url rule to find a relevant authenticator for it.

        :rtype: str
        """

        authenticator = self.get_rule_based_authenticator_name(url)
        if not authenticator:
            authenticator = self._default_authenticator

        return authenticator

    def get_current_authenticator_name(self):
        """
        gets the authenticator name for current request from its matched url rule.

        it may return None if the current request does not match any url rule.
        it also returns None for routes which have `authenticated=False` in their
        definition.

        :rtype: str
        """

        current_request = session_services.get_current_request()
        if current_request.url_rule and isinstance(current_request.url_rule, ProtectedRoute):
            return current_request.url_rule.authenticator

        return None

    def get_current_authenticator(self):
        """
        gets the authenticator for current request from its matched url rule.

        it may return None if the current request does not match any url rule.
        it also returns None for routes which have `authenticated=False` in their
        definition.

        :rtype: AbstractAuthenticatorBase
        """

        name = self.get_current_authenticator_name()
        if name:
            return self.get_authenticator(name)

        return None

    def get_rule_based_authenticator_name(self, url):
        """
        gets the relevant authenticator name to the given url rule.

        it may return None if no matching rule is found.

        :param str url: url to get its matching authenticator name.

        :rtype: str
        """

        if url[-1] != '/':
            url = f'{url}/'

        for rule, name in self._rule_based_authenticators.items():
            if url.startswith(rule):
                return name

        return None

    def authenticator_exists(self, name):
        """
        gets a value indicating that an authenticator with given name exists.

        :param str name: authenticator name.

        :rtype: bool
        """

        return name in self._authenticators

    def get_authenticator(self, name, **options):
        """
        gets the authenticator with given name.

        :keyword str name: authenticator name.

        :raises AuthenticatorNotFoundError: authenticator not found error.

        :rtype: AbstractAuthenticatorBase
        """

        if name not in self._authenticators:
            raise AuthenticatorNotFoundError('Authenticator [{name}] not found.'
                                             .format(name=name))

        return self._authenticators[name]

    def register_authenticator(self, instance, **options):
        """
        registers a new authenticator or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an authenticator which is already registered.

        :param AbstractAuthenticatorBase instance: authenticator to be registered.
                                                   it must be an instance of
                                                   AbstractAuthenticatorBase.

        :keyword bool replace: specifies that if there is another registered
                               authenticator with the same name, replace it with
                               the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidAuthenticatorTypeError: invalid authenticator type error.
        :raises DuplicatedAuthenticatorError: duplicated authenticator error.
        """

        if not isinstance(instance, AbstractAuthenticatorBase):
            raise InvalidAuthenticatorTypeError('Input parameter [{instance}] is '
                                                'not an instance of [{base}].'
                                                .format(instance=instance,
                                                        base=AbstractAuthenticatorBase))

        if instance.name in self._authenticators:
            old_instance = self._authenticators.get(instance.name)
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedAuthenticatorError('There is another registered '
                                                   'authenticator [{old}] with name '
                                                   '[{name}] but "replace" option is not '
                                                   'set, so authenticator [{instance}] '
                                                   'could not be registered.'
                                                   .format(old=old_instance,
                                                           name=instance.name,
                                                           instance=instance))

            print_warning('Authenticator [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._authenticators[instance.name] = instance

    def validate_authenticators(self):
        """
        validates that all protected routes authenticators are present.

        :raises AuthenticatorNotFoundError: authenticator not found error.
        """

        app = application_services.get_current_app()
        for route in app.url_map.get_routes():
            if isinstance(route, ProtectedRoute) and \
                    not self.authenticator_exists(route.authenticator):
                raise AuthenticatorNotFoundError('Authenticator with name [{name}] on '
                                                 'view function [{endpoint}] does not '
                                                 'exist in registered authenticators.'
                                                 .format(name=route.authenticator,
                                                         endpoint=route.endpoint))

    def authenticate(self, request, **options):
        """
        authenticates given request and pushes the authenticated data into current request.

        if authentication fails, authenticated data will not be pushed into current request.

        :param CoreRequest request: request to be authenticated.

        :keyword str authenticator: authenticator name to be used. if not
                                    provided, current request authenticator
                                    will be used. if current request does
                                    not have an authenticator, nothing will
                                    be done.

        :raises AuthenticationFailedError: authentication failed error.
        """

        authenticator_name = options.get('authenticator')
        if not authenticator_name:
            authenticator_name = self.get_current_authenticator_name()

        if authenticator_name:
            authenticator = self.get_authenticator(authenticator_name)
            try:
                authenticator.authenticate(request, **options)
            except AuthenticationFailedError:
                raise
            except Exception as error:
                raise AuthenticationFailedError(error) from error

    def login(self, username, password, authenticator, **options):
        """
        logs in a user with given info using provided authenticator.

        it may return the required credentials if they must be returned to client.

        :param str username: username.
        :param str password: password.
        :param str authenticator: authenticator name to be used.

        :raises ValidationError: validation error.
        :raises AuthenticatorNotFoundError: authenticator not found error.

        :returns: required credentials.
        """

        data = validator_services.validate('authentication',
                                           username=username,
                                           password=password)

        authenticator = self.get_authenticator(authenticator, **options)
        return authenticator.login(data.username, data.password, **options)

    def logout(self, authenticator, **options):
        """
        logouts the current user and clears its relevant credentials.

        :param str authenticator: authenticator name to be used.

        :raises AuthenticatorNotFoundError: authenticator not found error.
        """

        authenticator = self.get_authenticator(authenticator, **options)
        return authenticator.logout(**options)
